# ------------------------------------------------------------------------------
#
#   Anton Synytsia
#   anton.synytsia@gmail.com
#
# ------------------------------------------------------------------------------

dir = File.dirname(__FILE__)
require File.join(dir, 'core.rb')

AMS::Lib.require('window.rb')

module AMS::Cursor

  class << self

    private

    def wins
      AMS::Windows
    end

    public

    # Modifies cursor's visibility state.
    # @param [Boolean] state +true+ to show, +false+ to hide.
    # @return [Boolean] Whether the visibility state was changed.
    def show(state = true)
      state = (state == true)
      return false if (state == visible?)
      count = 0
      while true
        count = wins::ShowCursor.call(state)
        break if ((count > -1) == state)
      end
      true
    end


    # Determines whether the cursor is visible.
    # @return [Boolean] +true+ if visible, +false+ if hidden.
    def visible?
      buffer = [20,0,0,0,0].pack('l*')
      wins::GetCursorInfo.call(buffer)
      info = buffer.unpack('l*')
      (info[1] == 1)
    end


    # Retrieves cursor position.
    # @param [Integer] mode
    #   +1+ : Retrieves coordinates relative to screen origin.
    #   +2+ : Retrieves coordinates relative to the viewport origin.
    #   +3+ : Retrieves coordinates relative to the window's client origin.
    # @return [Array] +[x,y]+
    def get_pos(mode = 1)
      buffer = 0.chr*8
      wins::GetCursorPos.call(buffer)
      pos = buffer.unpack('ll')
      if mode == 2
        origin = AMS::Sketchup::Viewport.origin
        pos[0] -= origin[0]
        pos[1] -= origin[1]
      elsif mode == 3
        buffer = 0.chr*16
        wins::GetClientRect.call(AMS::Sketchup::MainWindow.hwnd, buffer)
        crect = buffer.unpack('l*')
        pos[0] -= crect[0]
        pos[1] -= crect[1]
      end
      pos
    end


    # Moves cursor to a specified position.
    # @param [Integer] x
    # @param [Integer] y
    # @param [Integer] mode
    #   +1+ : Given coordinates are relative to screen origin.
    #   +2+ : Given coordinates are relative to the viewport origin.
    #   +3+ : Given coordinates are relative to the window's client origin.
    # @return [Boolean] +true+ if successful, +false+ if not.
    def set_pos(x,y, mode = 1)
      x = x.to_i
      y = y.to_i
      if mode == 2
        origin = AMS::Sketchup::Viewport.origin
        x += origin[0]
        y += origin[1]
      elsif mode == 3
        buffer = 0.chr*16
        wins::GetClientRect.call(AMS::Sketchup::MainWindow.hwnd, buffer)
        crect = buffer.unpack('l*')
        x += crect[0]
        y += crect[1]
      end
      wins::SetCursorPos.call(x,y) == 1
    end


    # Retrieves the coordinates of the upper-left and lower-right corners of
    # a clip rectangle.
    # @param [Integer] mode
    #   +1+ : Retrieves coordinates relative to screen origin.
    #   +2+ : Retrieves coordinates relative to the viewport origin.
    #   +3+ : Retrieves coordinates relative to the window's client origin.
    # @return [Array] +[x1,y1, x2,y2]+
    def get_clip_rect(mode = 1)
      buffer = 0.chr*16
      wins.getClipCursor(buffer)
      rect = buffer.unpack('l*')
      if mode == 2
        origin = AMS::Sketchup::Viewport.origin
        rect[0] -= origin[0]
        rect[1] -= origin[1]
        rect[2] -= origin[0]
        rect[3] -= origin[1]
      elsif mode == 3
        buffer = 0.chr*16
        wins::GetClientRect.call(AMS::Sketchup::MainWindow.hwnd, buffer)
        crect = buffer.unpack('l*')
        rect[0] -= crect[0]
        rect[1] -= crect[1]
        rect[2] -= crect[0]
        rect[3] -= crect[1]
      end
      rect
    end


    # Assigns new coordinates of the upper-left and lower-right corners of
    # a clip rectangle.
    # @param [Array] rect +[x1,y1, x2,y2]+
    # @param [Integer] mode
    #   +1+ : Given coordinates are relative to screen origin.
    #   +2+ : Given coordinates are relative to the viewport origin.
    #   +3+ : Given coordinates are relative to the window's client origin.
    # @return [Boolean] +true+ if successful, +false+ if not.
    def set_clip_rect(rect, mode = 1)
      if mode == 2
        origin = AMS::Sketchup::Viewport.origin
        rect[0] += origin[0]
        rect[1] += origin[1]
        rect[2] += origin[0]
        rect[3] += origin[1]
      elsif mode == 3
        buffer = 0.chr*16
        wins::GetClientRect.call(AMS::Sketchup::MainWindow.hwnd, buffer)
        crect = buffer.unpack('l*')
        rect[0] += crect[0]
        rect[1] += crect[1]
        rect[2] += crect[0]
        rect[3] += crect[1]
      end
      buffer = rect.pack('llll')
      wins::ClipCursor.call(buffer) == 1
    end


    # Clips cursor to the upper-left and lower-right corners of the main window's
    # client area.
    # @return [Boolean] +true+ if successful, +false+ if not.
    def clip_to_main_window
      buffer = 0.chr*16
      wins::GetClientRect.call(AMS::Sketchup::MainWindow.hwnd, buffer)
      crect = buffer.unpack('l*')
      set_clip_rect(crect)
    end


    # Clips cursor to the upper-left and lower-right corners of the viewport
    # client area.
    # @return [Boolean] +true+ if successful, +false+ if not.
    def clip_to_viewport
      rect = AMS::Sketchup::Viewport.rect
      set_clip_rect(rect)
    end


    # Determines whether the cursor is pointing at the main window.
    # @return [Boolean]
    def main_window_target?
      buffer = 0.chr*8
      wins::GetCursorPos.call(buffer)
      x,y = buffer.unpack('ll')
      hwnd = wins::WindowFromPoint.call(x,y)
      main_hwnd = AMS::Sketchup.main_window
      (hwnd == main_hwnd) or (wins::IsChild.call(main_hwnd, hwnd) == 1)
    end


    # Determines whether the cursor is inside the viewport client area.
    # @return [Boolean]
    def viewport_target?
      buffer = 0.chr*8
      wins::GetCursorPos.call(buffer)
      x,y = buffer.unpack('ll')
      r = AMS::Window.client_rect(AMS::Sketchup.viewport)
      x.between?(r[0], r[2]) and y.between?(r[1], r[3])
    end

  end # proxy class

end # module AMS::Cursor

AMS::Lib.file_loaded(__FILE__)
