# ------------------------------------------------------------------------------
#
#   Anton Synytsia
#   anton.synytsia@gmail.com
#
# ------------------------------------------------------------------------------

dir = File.dirname(__FILE__)
require File.join(dir, 'core.rb')

AMS::Lib.require 'sketchup_window.rb'

# Input procedure is a wrap-up utility of the MSDN hook procedure. A hook is a
# point in the system message-handling mechanism where an application can
# install a subroutine to monitor the message traffic in the system and process
# certain types of messages before they reach the target window procedure. One
# of the most inspiring features of the hook procedure is monitoring keyboard
# and mouse messages that reach SketchUp window. This gives us an access to SU's
# message procedure. We can actually know when a certain input event occurs, and
# process it with our own code. Not only it is capable of monitoring messages,
# the hook procedure can also prevent the system from passing the message to
# the target window procedure. This allows us to block certain keyboard and
# mouse events as if they were not present in SketchUp at the first place.
# Receiving input of certain key strokes and mouse clicks without the presence
# of SketchUp shortcuts and mouse operations becomes a huge advantage for the
# FPS type gaming.
module AMS::InputProc

  class << self

    private

    def api
      AMS::Win32::API
    end

    def wins
      AMS::Windows
    end

  end # proxy class


  @hmod = AMS::Sketchup.module_handle
  @tid = AMS::Sketchup.thread_id
  @tool = nil
  @opts = []
  @hhooks = []
  @keyboard_methods = { :hk_onKeyDown => 0, :hk_onKeyUp => -1, :hk_onKeyExtended => 1 }
  @mouse_methods = {
    :hk_onMouseMove => 0x0200,
    :hk_onLButtonDown => 0x0201, :hk_onLButtonUp => 0x0202, :hk_onLButtonDoubleClick => 0x0203,
    :hk_onRButtonDown => 0x0204, :hk_onRButtonUp => 0x0205, :hk_onRButtonDoubleClick => 0x0206,
    :hk_onMButtonDown => 0x0207, :hk_onMButtonUp => 0x0208, :hk_onMButtonDoubleClick => 0x0209,
    :hk_onMouseWheelRotate => 0x020A,
    :hk_onXButtonDown => 0x020B, :hk_onXButtonUp => 0x020C, :hk_onXButtonDoubleClick => 0x020D,
    :hk_onMouseWheelTilt => 0x020E
  }
  # Stores an array of keyboard method values that are used by the selected
  # tool.
  @keyboard_use = []
  # Stores an array of mouse method values that are used by the selected tool.
  @mouse_use = []
  # This variable will be used record mouse wheel rotate or tilt direction from
  # the low-level mouse procedure. The standard or local mouse procedure
  # does not specify the direction, while the low-level does.
  @rt_dir = 0
  # For some reason SU receives F1 key's input despite that the local keyboard
  # procedure is preventing the system from passing the message. To avoid
  # getting the help dialog we will handle F1 key in the low level keyboard
  # procedure; yet, the low level procedure uses the same flag for the onKeyDown
  # and onKeyExtended event. Create a variable to record the last flag, and use
  # it to identify the new flag.
  @f1_state = -1
  # When preventing a system message the message notifications are not saved,
  # thus the double click events are not present. To make sure they work we
  # need to record the time of mouse button clicks and make decisions whether
  # the next click should be a double-click message.
  @click_rec = {}
  @last_down = nil
  @last_up = nil
  # In order to distinguish the click from the double-click message we must know
  # maximum double click delay and the rectangle in which the next click should
  # occur.
  @dc_time = wins::GetDoubleClickTime.call()*0.001
  @dc_rect = [wins::GetSystemMetrics.call(36), wins::GetSystemMetrics.call(37)]

  # Hook Procedure
  #   http://msdn.microsoft.com/en-us/library/windows/desktop/ms632589(v=vs.85).aspx


  # KeyboardProc callback function
  #   http://msdn.microsoft.com/en-us/library/windows/desktop/ms644984(v=vs.85).aspx
  #
  @keyboardProc = api::Callback.new('ILI', 'I'){ |nCode, wParam, lParam|
    if nCode != 0 or wParam == 0x0070
      next wins::CallNextHookEx.call(@hhooks[0], nCode, wParam, lParam)
    end
    val = 0
    state = lParam >> 30
    st1 = @keyboard_use.include?(state)
    st2 = @opts[1]
    st3 = AMS::Sketchup::MainWindow.active?
    if st1 and (st2 or st3)
      # Associate the message value with its name.
      func = RUBY_VERSION =~ /1.8/ ? :index : :key
      vk = wins::VirtualKeyCodes.method(func).call(wParam).to_s.downcase
      name = @keyboard_methods.method(func).call(state)
      # Call the function
      begin
        val = @tool.method(name).call(vk, wParam)
      rescue Exception => e
        val = 0
        puts "There was an error calling #{@tool}.#{name} method:\n#{e}\n#{$@[0..2].join("\n")}"
        select_tool nil
      end
    end
    # Pass the hook information to the next hook procedure.
    wins::CallNextHookEx.call(@hhooks[0], nCode, wParam, lParam)
    # May return a non-zero value to prevent the system from passing the message
    # to the target window procedure. In this case +true+ acts as a non-zero
    # value, while +false+ acts as a zero value.
    val == 1 ? 1 : 0
  }


  # LowLevelKeyboardProc callback function
  #   http://msdn.microsoft.com/en-us/library/windows/desktop/ms644985(v=vs.85).aspx
  #
  @keyboardProcLL = api::Callback.new('ILL', 'I'){ |nCode, wParam, lParam|
    if nCode != 0
      next wins::CallNextHookEx.call(@hhooks[1], nCode, wParam, lParam)
    end
    val = 0
    st1 = @opts[1]
    st2 = AMS::Sketchup.active?
    st3 = AMS::Sketchup::MainWindow.active?
    if (st1 and st2) or st3
      struct = 0.chr*4
      wins::Memcpy2.call(struct, lParam, struct.size)
      vk = struct.unpack('l')[0]
      if vk == 0x0070
        # Here we need to handle the F1 key to avoid getting the help dialog.
        # For some reason SU receives its input despite that the standard
        # KeyboardProc is preventing the system from passing the message to
        # the target window procedure. The flags for the +onKeyDown+ and
        # +onKeyExtened+ are the same. Use a variable to identify the flag.
        # onKeyDown      : 0x100 -> 0
        # onKeyExteneded : 0x100 -> 1
        # onKeyUp        : 0x101 -> -1
        @f1_state = ( wParam == 0x101 ? -1 : (@f1_state == -1 ? 0 : 1) )
        # Associate the message value with its method name.
        func = RUBY_VERSION =~ /1.8/ ? :index : :key
        name = @keyboard_methods.method(func).call(@f1_state)
        # Call the function
        begin
          val = @tool.method(name).call('f1', 0x70)
        rescue Exception => e
          val = 0
          puts "There was an error calling #{@tool}.#{name} method:\n#{e}\n#{$@[0..2].join("\n")}"
          select_tool nil
        end
      end
    end
    # Pass the hook information to the next hook procedure.
    wins::CallNextHookEx.call(@hhooks[1], nCode, wParam, lParam)
    # May return a non-zero value to prevent the system from passing the message
    # to the target window procedure. In this case +true+ acts as a non-zero
    # value, while +false+ acts as a zero value.
    val == 1 ? 1 : 0
  }


  # MouseProc callback function
  #   http://msdn.microsoft.com/en-us/library/windows/desktop/ms644988(v=vs.85).aspx
  #
  @mouseProc = api::Callback.new('ILL', 'I'){ |nCode, wParam, lParam|
    if nCode != 0 or !@mouse_use.include?(wParam)
      next wins::CallNextHookEx.call(@hhooks[2], nCode, wParam, lParam)
    end
    val = 0
    st1 = @opts[1]
    st2 = AMS::Sketchup::MainWindow.active?
    st3 = AMS::Cursor.main_window_target?
    st4 = wParam.between?(0x0201, 0x0209) or wParam.between?(0x020B, 0x020D)
    if st1 or st2 or (st3 and st4)
      struct = 0.chr*12
      wins::Memcpy2.call(struct, lParam, struct.size)
      x,y, hwnd = struct.unpack('l*')
      # Make x,y relative to the viewport origin
      if @opts[0]
        origin = AMS::Sketchup::Viewport.origin
        x -= origin[0]
        y -= origin[1]
      end
      # Associate the message value with its method name.
      func = RUBY_VERSION =~ /1.8/ ? :index : :key
      name = @mouse_methods.method(func).call(wParam)
      # Check if the current on[x]ButtonDown event must be a double-click
      # event. What makes a on[x]ButtonDown a double-click event:
      #   * The next on[x]ButtonDown must be within the maximum double-click
      #     delay and within the certain rectangular area.
      #   * An on[x]ButtonDown event cannot be double-click event if the last
      #     on[x]ButtonDown event was a double-click event.
      #   * There must be an on[x]ButtonUp in between the two on[x]ButtonDown
      #     events.
      if name =~ /Down/
        data = @click_rec[name]
        @click_rec[name] = [Time.now, x,y]
        if data != nil and name[5,1] == @last_up and !(@last_down =~ /DoubleClick/)
          dt = Time.now - data[0]
          dx = (x - data[1]).abs
          dy = (y - data[2]).abs
          status = ( dt <= @dc_time and dx <= @dc_rect[0] and dy <= @dc_rect[1] )
          name = "hk_on#{name[5,1]}ButtonDoubleClick" if status
        end
      elsif name =~ /DoubleClick/ and name[5,1] != @last_up
        name = "hk_on#{name[5,1]}ButtonDown"
      end
      @last_down = name if name =~ /Down/ or name =~ /DoubleClick/
      @last_up = name[5,1] if name =~ /Up/
      # Set the method parameters
      params = [x,y]
      params.push @rt_dir if [0x020A, 0x020E].include?(wParam)
      params.push hwnd if @opts[2]
      # Call the function
      begin
        val = @tool.method(name).call *params
      rescue Exception => e
        val = 0
        puts "There was an error calling #{@tool}.#{name} method:\n#{e}\n#{$@[0..2].join("\n")}"
        select_tool nil
      end
    end
    # Pass the hook information to the next hook procedure.
    wins::CallNextHookEx.call(@hhooks[2], nCode, wParam, lParam)
    # May return a non-zero value to prevent the system from passing the message
    # to the target window procedure. In this case +true+ acts as a non-zero
    # value, while +false+ acts as a zero value.
    val == 1 ? 1 : 0
  }


  # LowLevelMouseProc callback function
  #   http://msdn.microsoft.com/en-us/library/windows/desktop/ms644986(v=vs.85).aspx
  #
  @mouseProcLL = api::Callback.new('ILL', 'I'){ |nCode, wParam, lParam|
    if nCode == 0 and [0x020A, 0x020E].include?(wParam) and AMS::Sketchup.active?
      # Use MouseProcLL to get mouse wheel rotate and tilt directions, since the
      # standard MouseProc does not state them.
      struct = 0.chr*12
      wins::Memcpy2.call(struct, lParam, struct.size)
      x,y, mouseData = struct.unpack('l*')
      @rt_dir = ( mouseData <=> 0 )
    end
    # Pass the hook information to the next hook procedure.
    wins::CallNextHookEx.call(@hhooks[3], nCode, wParam, lParam)
  }


  class << self

    # Selects an object for the input procedure. The object must be a module,
    # class, or a class instance that will contain one or more of the following
    # callback methods below. Pass +nil+ to deselect the current tool. When a
    # new tool is selected, the original tool is deactivated, and the new tool
    # is activated. You may also pass the same tool and it will behave the same
    # way as if the new tool is selected; the tool is deactivated, and then
    # activated once again. This behaves pretty much the same way as the
    # standard Sketchup tool. A special _hk_ extension is added to each of the
    # callback functions below to make them distinguishable and compatible when
    # they are mixed with the standard SketchUp tool callback functions.
    #
    # The tool must contain one or more of the following callback methods:
    #   hk_activate
    #   hk_deactivate
    #   hk_onKeyDown(key, val)
    #   hk_onKeyUp(key, val)
    #   hk_onKeyExtended(key, val)
    #   hk_onMouseMove(x,y)
    #   hk_onLButtonDown(x,y)
    #   hk_onLButtonUp(x,y)
    #   hk_onLButtonDoubleClick(x,y)
    #   hk_onRButtonDown(x,y)
    #   hk_onRButtonUp(x,y)
    #   hk_onRButtonDoubleClick(x,y)
    #   hk_onMButtonDown(x,y)
    #   hk_onMButtonUp(x,y)
    #   hk_onMButtonDoubleClick(x,y)
    #   hk_onXButtonDown(x,y)
    #   hk_onXButtonUp(x,y)
    #   hk_onXButtonDoubleClick(x,y)
    #   hk_onMouseWheelRotate(x,y, dir)
    #   hk_onMouseWheelTilt(x,y, dir)
    #
    # Each of these callback functions (except +hk_activate+ and
    # +hk_deactivate+) may return a value of +1+ to prevent the system from
    # passing message to the SketchUp window or its thread, depending on the
    # +opt2+ parameter. Passing a return value of +1+ to the +hk_onMouseMove+
    # will not stop the mouse movement, but it will prevent all of the mouse
    # notifications being sent to SketchUp, meaning +onMouseMove+,
    # +onMouseEnter+, and +onMouseLeave+ callback functions of the standard
    # SketchUp tool will not be called. Other hook methods work with the same
    # principle. For example, setting a return value of +1+ to the
    # +hk_onLButtonDown+ will prevent the message from reaching SketchUp's
    # window procedure, thus the +onLButtonDown+ callback function of the
    # standard SketchUp tool will not be called! See
    # <i>examples/ams_InputProc.rb</i> for detailed implementation.
    #
    # @param [Object, nil] tool Pass +nil+ to deselect the current selected
    #   tool, pass +module/class/instance+ to set the new selected tool.
    # @param [Boolean] opt1 Pass +true+ if want to have cursor position
    #   parameters relative to the viewport origin, +false+ if relative to the
    #   screen origin.
    # @param [Boolean] opt2 +true+ if want to hook the whole SU thread
    #   (toolbars and dialogues), +false+ if just want to hook the main window.
    # @param [Boolean] opt3 +true+ if want to have an extra hwnd parameter to
    #   each the mouse callback functions, +false+ if not. The hwnd will
    #   specify the handle to the window the cursor is currently pointing at.
    # @return [Boolean] +true+ if success, +false+ if the selected tool doesn't
    #   use any of the callback methods above (regarding +hk_activate+ and
    #   +hk_deactivate+), or there's an error in +hk_activate+ callback method.
    # @see AMS::InputProc::Tool
    def select_tool(tool, opt1 = true, opt2 = false, opt3 = false)
      # Unhook the original tool
      if @tool
        @hhooks.each{ |hwnd| wins::UnhookWindowsHookEx.call(hwnd) }
        @hhooks.clear
        method = :hk_deactivate
        method = method.to_s if RUBY_VERSION.to_f < 1.9
        begin
          @tool.method(method).call
        rescue Exception => e
          puts "There was an error calling #{@tool}.#{method} method:\n#{e}\n#{$@[0..2].join("\n")}"
        end if @tool.methods.include?(method)
      end
      # Reset variables
      @tool = nil
      @keyboard_use.clear
      @mouse_use.clear
      @f1_state = -1
      @click_rec.clear
      @opts.clear
      return true if tool == nil
      methods = tool.methods
      # Convert strings to symbols
      for n in 0...methods.size
        methods[n] = methods[n].to_sym
      end if RUBY_VERSION =~ /1.8/
      # Verify the new tool
      keyboard_use = []
      mouse_use = []
      (methods & @keyboard_methods.keys).each{ |method|
        keyboard_use.push @keyboard_methods[method]
      }
      (methods & @mouse_methods.keys).each{ |method|
        mouse_use.push @mouse_methods[method]
      }
      usage = keyboard_use.size + mouse_use.size
      return false if usage.zero?
      # Call the activate hook
      begin
        tool.hk_activate
      rescue Exception => e
        puts "There was an error calling #{tool}.hk_activate method:\n#{e}\n#{$@[0..2].join("\n")}"
        begin
          tool.hk_deactivate
        rescue Exception => e
          puts "There was an error calling #{tool}.hk_deactivate method:\n#{e}\n#{$@[0..2].join("\n")}"
        end if methods.include?(:hk_deactivate)
        return false
      end if methods.include?(:hk_activate)
      # Record info
      @keyboard_use = keyboard_use
      @mouse_use = mouse_use
      @opts = [(opt1 == true), (opt2 == true), (opt3 == true)]
      @tool = tool
      # Hook the new tool
      unless @keyboard_use.empty?
        @hhooks[0] = wins::SetWindowsHookEx.call(2, @keyboardProc, nil, @tid)
        @hhooks[1] = wins::SetWindowsHookEx.call(13, @keyboardProcLL, @hmod, nil)
      end
      unless @mouse_use.empty?
        @hhooks[2] = wins::SetWindowsHookEx.call(7, @mouseProc, nil, @tid)
        unless (@mouse_use & [0x020A, 0x020E]).empty?
          @hhooks[3] = wins::SetWindowsHookEx.call(14, @mouseProcLL, @hmod, nil)
        end
      end
      true
    end


    # Deselects the tool only if the specified tool is the current active tool.
    # @param [Object] tool
    # @return [Boolean] +true+ if successful, +false+ if not.
    def deselect_tool(tool)
      select_tool nil if tool == @tool
    end


    # Retrieves the selected tool.
    # @return [Object, nil]
    def selected_tool
      @tool
    end

  end # proxy class


  # @private
  class MyAppObserver < Sketchup::AppObserver

    # Ensures safe unhook.
    def onQuit
      ::AMS::InputProc.select_tool nil
    end

  end # class MyAppObserver

end # module AMS::InputProc


unless AMS::Lib.file_loaded?(__FILE__)
  app_obs = AMS::InputProc::MyAppObserver.new
  Sketchup.add_observer(app_obs)
  AMS::Lib.file_loaded(__FILE__)
end
