# ------------------------------------------------------------------------------
#
#   Anton Synytsia
#   anton.synytsia@gmail.com
#
# ------------------------------------------------------------------------------

dir = File.dirname(__FILE__)
require File.join(dir, 'core.rb')

AMS::Lib.require 'windows.rb'

module AMS::Keyboard

  class << self

    private

    def wins
      AMS::Windows
    end

    public

    # Retrieves state of the specified key.
    # @param [String, Symbol, Integer] vk Virtual key name or value.
    # @return [Integer] +1+ if down, +0+ if up.
    def key_state(vk)
      if [String, Symbol].include?(vk.class)
        vk = vk.to_s.downcase
        v = wins::VirtualKeyCodes[vk]
        v = wins::VirtualKeyCodes2[vk] if v == nil
        return 0 if v == nil
        vk = v
      end
      ( wins::GetKeyState.call(vk)%256 >> 7 )
    end

    alias key key_state

    # Retrieves constant value of the specified key.
    # @param [String, Symbol] vk Virtual key name.
    # @return [Integer] Virtual key value.
    def get_value(vk)
      vk = vk.to_s.downcase
      v = wins::VirtualKeyCodes[vk]
      v = wins::VirtualKeyCodes2[vk] if v.nil?
      return 0 if v.nil?
      v
    end

    # Checks whether the specified key is down.
    # @param [String, Symbol, Integer] vk Virtual key name or value.
    # @return [Boolean] +true+ if down, +false+ if up.
    def key_down?(vk)
      key(vk) == 1
    end


    # Checks whether the specified key is up.
    # @param [String, Symbol, Integer] vk Virtual key name or value.
    # @return [Boolean] +true+ if up, +false+ if down.
    def key_up?(vk)
      key(vk) == 0
    end


    # Checks whether an array of specified keys are down.
    # @param [Array<String, Symbol, Integer>] keys An array of virtual key names
    #   or values.
    # @return [Boolean] +true+ if all are down, +false+ if some or all are up.
    def keys_down?(*keys)
      keys.flatten!
      return false if keys.empty?
      keys.each{ |vk| return false if (key(vk) == 0) }
      true
    end


    # Checks whether an array of specified keys are up.
    # @param [Array<String, Symbol, Integer>] keys An array of virtual key names
    #   or values.
    # @return [Boolean] +true+ if all are up, +false+ if some or all are down.
    def keys_up?(*keys)
      keys.flatten!
      return false if keys.empty?
      keys.each{ |vk| return false if (key(vk) == 1) }
      true
    end


    # Retrieves an array of states to all 254 keys.
    # @return [Array<Integer>] An array of 256 values of +1+ (down) or +0+ (up).
    def keyboard_state
      buffer = 0.chr*256
      wins::GetKeyboardState.call(buffer)
      x = buffer.unpack('C*')
      for n in 0...256; x[n] = x[n] >> 7 end
      x
    end

  end # proxy class

end # module AMS::Keyboard

AMS::Lib.file_loaded(__FILE__)
