# ------------------------------------------------------------------------------
#
#   Anton Synytsia
#   anton.synytsia@gmail.com
#
# ------------------------------------------------------------------------------

dir = File.dirname(__FILE__)
require File.join(dir, 'core.rb')

# The ray utility adds more options to the +Sketchup.active_model.raytest+
# method.
module AMS::RayUtil

  module_function

  # Retrieves an array of points intersecting the ray.
  # _T1_ : type 1 checks all model entities.
  # @param [Geom::Point3d, Array] point Ray position.
  # @param [Geom::Vector3d, Array] vector Ray direction.
  # @param [Boolean] chg Whether to consider hidden geometry.
  # @return [Array<Geom::Point3d>] An array of points intersecting the ray.
  def deepray_t1(point, vector, chg = false)
    chg = chg ? true : false
    pts = []
    result = nil
    while true
      result = Sketchup.active_model.raytest(point, vector, !chg)
      break unless result
      x = result[0]
      pts.push x
      point = x
    end
    pts
  end


  # Retrieves an array of points intersecting the ray.
  # _T2_ : type 2 checks all the given entities.
  # @param [Array<entity>] ents An array of entities to include.
  # @param [Geom::Point3d, Array] point Ray position.
  # @param [Geom::Vector3d, Array] vector Ray direction.
  # @param [Boolean] chg Whether to consider hidden geometry.
  # @return [Array<Geom::Point3d>] An array of points intersecting the ray.
  def deepray_t2(ents, point, vector, chg = false)
    chg = chg ? true : false
    ents = [ents] unless ents.is_a?(Array)
    pts = []
    result = nil
    while true
      result = Sketchup.active_model.raytest(point, vector, !chg)
      break unless result
      x = result[0]
      pts.push x if ents.include?(result[1][0])
      point = x
    end
    pts
  end


  # Retrieves an array of points intersecting the ray.
  # _T3_ : type 3 checks all, but the given entities.
  # @param [Array<entity>] ents An array of entities to ignore.
  # @param [Geom::Point3d, Array] point Ray position.
  # @param [Geom::Vector3d, Array] vector Ray direction.
  # @param [Boolean] chg Whether to consider hidden geometry.
  # @return [Array<Geom::Point3d>] An array of points intersecting the ray.
  def deepray_t3(ents, point, vector, chg = false)
    chg = chg ? true : false
    ents = [ents] unless ents.is_a?(Array)
    pts = []
    result = nil
    while true
      result = Sketchup.active_model.raytest(point, vector, !chg)
      break unless result
      x = result[0]
      pts.push x unless ents.include?(result[1][0])
      point = x
    end
    pts
  end


  # Casts a ray through the model and return the first thing that the ray hits.
  # _T1_ : type 1 checks all given entities.
  # @param [Array<entity>] ents An array of entities to include.
  # @param [Geom::Point3d, Array] point Ray position.
  # @param [Geom::Vector3d, Array] vector Ray direction.
  # @param [Boolean] chg Whether to consider hidden geometry.
  # @return [Array, nil] A ray result:
  #   http://www.sketchup.com/intl/en/developer/docs/ourdoc/model.php#raytest
  def raytest_t1(ents, point, vector, chg = false)
    chg = chg ? true : false
    ents = [ents] unless ents.is_a?(Array)
    result = nil
    while true
      result = Sketchup.active_model.raytest(point, vector, !chg)
      break unless result
      return result if ents.include?(result[1][0])
      point = result[0]
    end
    nil
  end


  # Casts a ray through the model and return the first thing that the ray hits.
  # _T2_ : type 2 checks all, but the given entities.
  # @param [Array<entity>] ents An array of entities to ignore.
  # @param [Geom::Point3d, Array] point Ray position.
  # @param [Geom::Vector3d, Array] vector Ray direction.
  # @param [Boolean] chg Whether to consider hidden geometry.
  # @return [Array, nil] A ray result:
  #   http://www.sketchup.com/intl/en/developer/docs/ourdoc/model.php#raytest
  def raytest_t2(ents, point, vector, chg = false)
    chg = chg ? true : false
    ents = [ents] unless ents.is_a?(Array)
    result = nil
    while true
      result = Sketchup.active_model.raytest(point, vector, !chg)
      break unless result
      return result unless ents.include?(result[1][0])
      point = result[0]
    end
    nil
  end

  # Casts a ray through the model and return the first thing that the ray hits.
  # _T3_ : type 3 passes through transparent faces and stops until it hits a
  # solid face.
  # @param [Geom::Point3d, Array] point Ray position.
  # @param [Geom::Vector3d, Array] vector Ray direction.
  # @param [Boolean] chg Whether to consider hidden geometry.
  # @return [Array, nil] A ray result:
  #   http://www.sketchup.com/intl/en/developer/docs/ourdoc/model.php#raytest
  def raytest_t3(point, vector, chg = false)
    chg = chg ? true : false
    result = nil
    model = Sketchup.active_model
    cam = model.active_view.camera
    while true
      result = model.raytest(point, vector, !chg)
      break unless result
      ent = result[1].last
      return result unless ent.is_a?(Sketchup::Face)
      angle = ent.normal.angle_between(cam.direction)
      mat = nil
      normal = ent.normal
      for i in 0...(result[1].size-1)
        e = result[1][i]
        normal.transform!(e.transformation)
        mat = e.material if e.material != nil
      end
      fmat = angle < 90.degrees ? ent.back_material : ent.material
      mat = fmat if fmat != nil
      return result if mat.nil? or mat.alpha == 1.0
      point = result[0]
    end
    nil
  end

end # module AMS::RayUtil

AMS::Lib.file_loaded(__FILE__)
