# ------------------------------------------------------------------------------
#
#   Anton Synytsia
#   anton.synytsia@gmail.com
#
# ------------------------------------------------------------------------------

Sketchup.require 'ams_Lib/sketchup_window.rb'

# Main window is SketchUp's top, parent window.
module AMS::Sketchup::MainWindow

  @exstyle = 0x00000100
  @style = 0x14cfc000
  @placement = 0

  class << self

    private

    def wins
      AMS::Windows
    end

    public

    # Retrieves a handle to the main window.
    # @return [Integer] hWnd
    def hwnd
      AMS::Sketchup.main_window
    end

    alias handle hwnd


    # Retrieves window's class name.
    # @return [String, nil] Window's class name.
    def class_name
      return nil unless hwnd
      cname = 0.chr*64
      wins::GetClassName.call(hwnd, cname, cname.size)
      cname.strip
    end


    # Retrieves window's title text.
    # @param [Long] hwnd A handle to the window.
    # @return [String, nil] caption
    def caption
      AMS::Window.get_caption(hwnd)
    end


    # Modifies window's title text.
    # @param [String] text New caption.
    # @return [Boolean] Whether the text was changed.
    def caption=(text)
      state = AMS::Window.set_caption(hwnd, text)
      state
    end


    # Determines whether the main window is the window with which the user is
    # currently working.
    # @return [Boolean]
    def foreground?
      wins::GetForegroundWindow.call() == hwnd
    end

    alias active? foreground?


    # Determines whether the window is maximized.
    # @return [Boolean]
    def maximized?
      return false if fullscreen?
      wins::IsZoomed.call(hwnd) == 1
    end


    # Determines whether the window is minimized.
    # @return [Boolean]
    def minimized?
      return false if fullscreen?
      wins::IsIconic.call(hwnd) == 1
    end


    # Determines whether the window is restored (not maximized, nor minimized).
    # @return [Boolean]
    def restored?
      return false if fullscreen?
      !(maximized? or minimized?)
    end


    # Determines whether the window is fullscreen.
    # @return [Boolean] status
    def fullscreen?
      style = wins::GetWindowLong.call(hwnd, -16)
      [0x3483c000, 0x3583c000, 0x1483c000, 0x1583c000].include?(style)
    end


    # Brings the thread that created the window into the foreground and
    # activates the main window.
    # @return [Boolean] +true+ if the window was brought to the foreground.
    def set_foreground
      wins::SetForegroundWindow.call(hwnd) == 1
    end

    alias set_active set_foreground


    # Sets main window maximized.
    # @return [Boolean] Whether window's state was changed.
    def maximize
      if fullscreen?
        set_fullscreen(false, 1)
      else
        return false if maximized?
        wins::ShowWindow.call(hwnd, 3)
      end
      true
    end


    # Sets main window minimized.
    # @return [Boolean] Whether window's state was changed.
    def minimize
      return false if minimized?
      wins::ShowWindow.call(hwnd, 2)
      true
    end


    # Sets main window to default placement.
    # @return [Boolean] Whether window's state was changed.
    def restore
      if fullscreen?
        set_fullscreen(false, 0)
      else
        return false if restored?
        wins::ShowWindow.call(hwnd, 1)
      end
      true
    end


    # Sets main window full screen.
    # @param [Boolean] state
    #   +true+ : set full screen,
    #   +false+ : set back to original placement.
    # @param [Integer] mode Whether to set normal (0), set maximized (1), or set
    #   to original placement (2). This option is used when main window is unset
    #   full screen.
    # @param [Boolean] show_menubar Whether to show/hide menubar when the window
    #   is set full screen.
    # @return [Boolean] Whether window's state was changed.
    def set_fullscreen(state = true, mode = 2, show_menubar = true)
      state = (state == true)
      show_menubar = (show_menubar == true)
      return false if state == fullscreen?
      wins::ShowWindow.call(hwnd, 1) if minimized?
      if state
        # Save original styles and placement
        style = wins::GetWindowLong.call(hwnd, -16)
        @style = 0x14cfc000 | (style & 0x01000000)
        @placement = 0.chr*44
        wins::GetWindowPlacement.call(hwnd, @placement)
        # Set full screen styles and placement
        exstyle = 0
        style = 0x0583c000 # ( WS_CLIPSIBLINGS | WS_MAXIMIZE | WS_TABSTOP | WS_GROUP | SS_WORDELLIPSIS )
        flags = 0x0664 # ( SWP_NOOWNERZORDER | SWP_NOSENDCHANGING | SWP_SHOWWINDOW | SWP_NOZORDER | SWP_FRAMECHANGED )
      else
        # Set original styles and placement
        exstyle = @exstyle
        style = @style
        pl = @placement.unpack('LLLllllllll')
        case mode
        when 0
          style = @style & ~0x01000000
          pl[1..2] = [0, 1]
        when 1
          style = @style | 0x01000000
          pl[1..2] = [2, 3]
        end
        @placement = pl.pack('LLLllllllll')
        flags = 0x0667 # ( SWP_NOOWNERZORDER | SWP_NOSENDCHANGING | SWP_SHOWWINDOW | SWP_NOZORDER | SWP_FRAMECHANGED | SWP_NOSIZE | SWP_NOMOVE )
        wins::SetWindowPlacement.call(hwnd, @placement)
      end
      # Set settings
      wins::LockWindowUpdate.call hwnd
      wins::SetWindowLong.call(hwnd, -20, exstyle)
      wins::SetWindowLong.call(hwnd, -16, style)
      wins::LockWindowUpdate.call nil
      w = wins::GetSystemMetrics.call(0) # SM_CXSCREEN
      h = wins::GetSystemMetrics.call(1) # SM_CYSCREEN
      mh = wins::GetSystemMetrics.call(15) # SM_CYMENU
      x = 0
      y = 0
      if (AMS::Sketchup::Menubar.set? and !show_menubar)
        y -= mh
        h += mh
      end
      wins::SetWindowPos.call(hwnd, nil, x, y, w, h, flags)
      true
    end


    # Changes the position and dimensions of the main window. The coordinates
    # must be specified relative to the upper-left corner of the screen.
    # @param [Integer] x
    # @param [Integer] y
    # @param [Integer] width
    # @param [Integer] height
    # @return [Boolean] +true+ if the function succeeds.
    def move(x = 100, y = 100, width = 800, height = 600)
      wins::MoveWindow.call(hwnd, x, y, width, height, true) == 1
    end

  end # proxy class

end # module AMS::Sketchup::MainWindow

AMS::Lib.file_loaded(__FILE__)
