# ------------------------------------------------------------------------------
#
#   Anton Synytsia
#   anton.synytsia@gmail.com
#
# ------------------------------------------------------------------------------

Sketchup.require 'ams_Lib/sketchup_window.rb'

# Menubar is a strip with SU commands. It is located below the window's caption.
module AMS::Sketchup::Menubar

  @commands = {}
  @shortcuts = {}
  @reg_shortcuts = {}
  @activation_delay = 0.2
  @remove_time = Time.new
  @csa =  false
  @locale = ::LanguageHandler.new('Localizable.strings')
  @locale2 = {}

  class << self

    private

    def wins
      AMS::Windows
    end

    public

    attr_reader :locale, :locale2

    # Retrieves a handle to the main window's menu-bar.
    # @return [Integer] hMenu
    def hmenu
      AMS::Sketchup.menubar
    end

    alias handle hmenu


    # Sets or removes menubar from the main window.
    # @param [Boolean] state +true+ : set, +false+ : remove.
    # @return [Boolean] Whether the menubar state was changed.
    def set(state = true)
      state = (state == true)
      hMenu = (state ? hmenu : nil)
      return false if (state == set?)
      unless state
        @remove_time = Time.now
        @csa = false
        @shortcuts = get_menu_shortcuts
        @reg_shortcuts = get_registry_shortcuts
        show(true)
      end
      wins::EndMenu.call()
      wins::SetMenu.call(AMS::Sketchup.main_window, hMenu)
      wins::DrawMenuBar.call(AMS::Sketchup.main_window)
      true
    end


    # Determines whether the current menubar is set.
    # @return [Boolean] Whether the current menubar is there.
    def set?
      wins::GetMenu.call(AMS::Sketchup.main_window) == hmenu
    end


    # Shows or hides menu bar of the main window. This function sizes the main
    # window that way, so that menu bar appears above the screen.
    # @note This can be done when the main window is full screen only. This also
    #   requires the menu bar to set.
    # @note You'll still be able to access menu by pressing the +Alt+ key.
    # @param [Boolean] state +true+ : show, +false+ : hide.
    # @return [Boolean] +true+ if +successful+ and visibility state changed,
    #   +false+ if not.
    def show(state = true)
      state = (state == true)
      return false unless set?
      return false unless AMS::Sketchup::MainWindow.fullscreen?
      return false if state == visible?
      main = AMS::Sketchup.main_window
      w = wins::GetSystemMetrics.call(0) # SM_CXSCREEN
      h = wins::GetSystemMetrics.call(1) # SM_CYSCREEN
      mh = wins::GetSystemMetrics.call(15) # SM_CYMENU
      if state
        wins::SetWindowPos.call(main, nil, 0, 0, w, h, 0x0644)
      else
        wins::SetWindowPos.call(main, nil, 0, -mh, w, h+mh, 0x0644)
      end
      true
    end


    # Determines whether menubar of the main window is visible.
    # @return [Boolean] +true+ : visible, +false+ : hidden or removed.
    def visible?
      return false unless set?
      return true unless AMS::Sketchup::MainWindow.fullscreen?
      main = AMS::Sketchup.main_window
      mh = wins::GetSystemMetrics.call(15) # SM_CYMENU
      rect = AMS::Window.rect(main)
      rect[1] > -mh
    end


    # Retrieves a hash of menu paths with their identifications. You may use
    # +Sketchup.send_action+ to perform any menu operation.
    # @param [Long] menu A handle to the menu.
    # @param [String] cur_path Current path of the menu.
    # @return [Hash]
    # @example
    #   commands = AMS::Menubar.get_menu_commands
    #   id = commands['Edit/Unhide/All']
    #   Sketchup.send_action(id)
    def get_menu_commands(menu = hmenu, cur_path = '')
      if RUBY_VERSION.to_f > 1.8
        Encoding.default_external = Encoding::UTF_8
        Encoding.default_internal = Encoding::UTF_8
      end
      return {} if wins::IsMenu.call(menu).zero?
      cur_path = cur_path.to_s
      data = {}
      count = wins::GetMenuItemCount.call(menu)
      data[cur_path] = hmenu
      for i in 0...count
        text = 0.chr*128
        size = wins::GetMenuString.call(menu, i, text, text.size, 0x0400)
        next if size.zero?
        name = text[0, size].unpack("C*").pack("U*").split("\t")[0].delete("&")
        path = cur_path.empty? ? name : cur_path + '/' + name
        id = wins::GetMenuItemID.call(menu, i)
        if (id == -1)
          subMenu = wins::GetSubMenu.call(menu, i)
          data.merge! get_menu_commands(subMenu, path)
        else
          data[path] = id
        end
      end
      data
    end


    # Retrieves menu identifier of the specified menu path.
    # @param [String] path Menu path.
    # @return [Integer, NilClass]
    def get_menu_id_by_path(path)
      @commands[path.to_s]
    end


    # Retrieves menu path of the specified menu identifier.
    # @param [Integer] id Menu identifier.
    # @return [String, NilClass]
    def get_menu_path_by_id(id)
      func = RUBY_VERSION.to_f < 1.9 ? :index : :key
      @commands.method(func).call(id.to_i)
    end


    # Retrieves menu text of the specified menu identifier.
    # @param [Integer] id Menu identifier.
    # @return [String, NilClass]
    def get_menu_string_by_id(id)
      func = RUBY_VERSION.to_f < 1.9 ? :index : :key
      path = @commands.method(func).call(id.to_i)
      return unless path
      sub_menu = path.split('/')[0..-2].join('/')
      menu = @commands[sub_menu]
      text = 0.chr*128
      size = wins::GetMenuString.call(menu, id, text, text.size, 0)
      return if size.zero?
      text[0, size].unpack("C*").pack("U*").split("\t")[0].delete("&")
    end


    # Assigns menu text to the specified menu identifier.
    # @param [Integer] id Menu identifier.
    # @param [String] text New string.
    # @return [Boolean]
    def set_menu_string_by_id(id, text)
      func = RUBY_VERSION.to_f < 1.9 ? :index : :key
      path = @commands.method(func).call(id.to_i)
      return false unless path
      sub_menu = path.split('/')[0..-2].join('/')
      menu = @commands[sub_menu]
      @commands.delete(path)
      @commands[sub_menu+'/'+text.to_s] = id
      wins::ModifyMenu.call(menu, id, 0, id, text.to_s.unpack("U*").pack("C*"))
      wins::DrawMenuBar.call(::AMS::Sketchup.main_window)
      true
    end


    # Retrieves menu text of the specified menu position.
    # @param [Integer] pos Menu position.
    # @param [Long] menu A handle to the menu or submenu.
    # @return [String, NilClass]
    def get_menu_string_by_pos(pos, menu = hmenu)
      text = 0.chr*128
      size = wins::GetMenuString.call(menu, pos.to_i, text, text.size, 0x0400)
      return nil if size == 0
      text[0, size].unpack("C*").pack("U*").split("\t")[0].delete("&")
    end


    # Assigns menu text to the specified menu position.
    # @param [Integer] pos Menu position.
    # @param [String] text New string.
    # @param [Long] menu A handle to the menu or submenu.
    # @return [Boolean]
    def set_menu_string_by_pos(pos, text, menu = hmenu)
      id = wins::GetMenuItemID.call(menu, pos)
      wins::ModifyMenu.call(menu, pos, 0x0400, id, text.to_s.unpack("U*").pack("C*"))
    end


    # @visibility private
    def update_menu_commands
      @commands = get_menu_commands
      @locale2['Window'] = get_menu_string_by_pos(6)
    end

    # @visibility private
    def get_registry_shortcuts
      shortcuts = {}
      reg = AMS::Sketchup::Registry
      num = reg.read("Settings/Num_Shortcuts").to_i
      for i in 1..num
        data = reg.read("Settings/Shortcut_#{i}")
        next if data.nil?
        data = data.split(' ', 5)
        next if data.size != 5
        k = ''
        k += 'Ctrl+' if data[0].to_i == 1
        k += 'Alt+' if data[1].to_i == 1
        k += 'Shift+' if data[2].to_i == 1
        k += data[3]
        v = data[4]
        shortcuts[k] = v
      end
      shortcuts
    end

    # @visibility private
    def get_menu_shortcuts
      shortcuts = {}
      Sketchup.get_shortcuts.each{ |data|
        shortcut, path = data.split("\t")
        # Translate shortcut to English.
        keys = shortcut.split("+")
        for i in 0...keys.size
          func = RUBY_VERSION.to_f < 1.9 ? :index : :key
          str = @locale.instance_variable_get("@strings").method(func).call(keys[i])
          keys[i] = str if str
        end
        shortcut = keys.join("+")
        shortcuts[shortcut] = path.to_s
      }
      shortcuts
    end

    # @visibility private
    def get_active_shortcut(key, val)
      kbrd = AMS::Keyboard
      @shortcuts.keys.each{ |s|
        keys = s.split('+')
        next if kbrd.get_value(keys.last) != val
        k = 'Shift'
        next if (keys.include?(k) ? kbrd.key_up?(k) : kbrd.key_down?(k))
        k = 'Ctrl'
        next if (keys.include?(k) ? kbrd.key_up?(k) : kbrd.key_down?(k))
        k = 'Alt'
        next if (keys.include?(k) ? kbrd.key_up?(k) : kbrd.key_down?(k))
        func = RUBY_VERSION.to_f < 1.9 ? :index : :key
        return s
      }
      nil
    end

    # @visibility private
    def process_key(key, val)
      if set?
        @csa = false
        return
      end
      # This ensures that shortcuts are not called twice in some cases.
      return if (Time.now - @remove_time) <= @activation_delay
      s = get_active_shortcut(key, val)
      return if s.nil?
      unless @csa
        cmd = @reg_shortcuts[s]
        return if cmd.nil? or !cmd.include?('/')
        if Sketchup.get_locale == 'en-US'
          return if (cmd =~ /Edit\/Make Component/i or cmd =~ /View\/Edge Style\/Back Edges/i)
        end
        @csa = true
      end
      path = @shortcuts[s]
      id = @commands[path]
      Sketchup.send_action(id) if id
    end

    # @visibility private
    def swo_mw_onKeyDown(key, val)
      process_key(key, val)
    end

    # @visibility private
    def swo_mw_onKeyExtended(key, val)
      process_key(key, val)
    end

    # @visibility private
    def swo_mw_onMinimize
      id = get_menu_id_by_path("#{@locale2['Window']}/#{@locale.GetString('Show Dialogs')}")
      id = get_menu_id_by_path("#{@locale2['Window']}/#{@locale.GetString('Hide Dialogs')}") unless id
      set_menu_string_by_id(id, @locale.GetString('Hide Dialogs')) if id
    end

    # @visibility private
    def swo_mw_onCommand(cmd)
      id = get_menu_id_by_path("#{@locale2['Window']}/#{@locale.GetString('Show Dialogs')}")
      id = get_menu_id_by_path("#{@locale2['Window']}/#{@locale.GetString('Hide Dialogs')}") unless id
      if id != nil and [id, 1<<16|id].include?(cmd)
        str = get_menu_string_by_id(id)
        str_show = @locale.GetString('Show Dialogs')
        str_hide = @locale.GetString('Hide Dialogs')
        str = (str == str_hide ? str_show : str_hide)
        set_menu_string_by_id(id, str)
      end
    end

    # @visibility private
    def swo_mw_onEnterMenu
      update_menu_commands
    end

    # @visibility private
    def activate_custom_menu
      AMS::Sketchup.add_observer(self)
      # Wait till all menu items are loaded.
      UI.start_timer(1, false){
        update_menu_commands
      }
    end

    # @visibility private
    def deactivate_custom_menu
      AMS::Sketchup.remove_observer(self)
    end

  end # proxy class

end # module AMS::Sketchup::MenuBar

AMS::Lib.file_loaded(__FILE__)
