# ------------------------------------------------------------------------------
#
#   Anton Synytsia
#   anton.synytsia@gmail.com
#
# ------------------------------------------------------------------------------

Sketchup.require 'ams_Lib/sketchup_window.rb'

module AMS::Sketchup::Registry

  class << self

    private

    def wins
      AMS::Windows
    end

    public

    def path
      PATH
    end


    # Retrieves path to SketchUp registry.
    # @return [String]
    def get_registry_path
      path = 'Software/'
      v = Sketchup.version.to_i
      path += case v
      when 5
        "Google/Google SketchUp5/"
      when 6..8
        "Google/SketchUp#{v}/"
      else
        "SketchUp/SketchUp 20#{v}/"
      end
      path
    end


    # @visibility private
    def format_path(path)
      data = path.split('/')
      if (data.size < 2)
        path = ""
        name = data[0].to_s
      else
        path = data[0...-1].join('/')
        name = data[-1]
      end
      fpath = File.join(PATH, path).gsub("/", "\\")
      # Encode to ASCII-8BIT
      fpath = fpath.unpack("U*").pack("C*")
      name = name.unpack("U*").pack("C*")
      [fpath, name]
    end


    # Retrieves a value to the specified registry key.
    # @param [String] path
    # @return [String, Fixnum, Bignum, NilClass] data, or +nil+ if path is
    #   invalid.
    def read(path)
      path, name = format_path(path)
      hKeyPtr = 0.chr*4
      r = wins::RegOpenKey.call(HKEY_CURRENT_USER, path, hKeyPtr)
      return if r != 0
      hKey = hKeyPtr.unpack('L')[0]
      typePtr = 0.chr*4
      dataPtr = 0.chr*2048
      sizePtr = [dataPtr.size].pack('L')
      r = wins::RegQueryValueEx.call(hKey, name, nil, typePtr, dataPtr, sizePtr)
      if r == 0
        type = typePtr.unpack('L')[0]
      else
        r = wins::RegQueryValue.call(hKey, name, dataPtr, sizePtr)
        type = 1
      end
      wins::RegCloseKey.call(hKey)
      return if r != 0
      size = sizePtr.unpack('L')[0]
      case type
      when 0 # REG_NONE
        nil
      when 1, 7 # REG_SZ, REG_MULTI_SZ
        dataPtr[0, size-1].unpack("C*").pack("U*")
      when 2 # REG_EXPAND_SZ
        buffer = 0.chr*2048
        s = wins::ExpandEnvironmentStrings.call(dataPtr, buffer, buffer.size)
        buffer[0, s-1].unpack("C*").pack("U*")
      when 3 # REG_BINARY
        dataPtr.unpack('A')[0]
      when 4 # REG_DWORD, REG_DWORD_LITTLE_ENDIAN
        dataPtr.unpack('L')[0]
      when 5 # REG_DWORD_BIG_ENDIAN
        dataPtr.unpack('q>')[0]
      when 6 # REG_LINK
        dataPtr.unpack('P')[0]
      when 11 # REG_QWORD, REG_QWORD_LITTLE_ENDIAN
        dataPtr.unpack('q')[0]
      end
    end

    alias :get :read


    # Sets/Creates/Replaces/Assigns a value to the specified registry path.
    # @param [String] path
    # @param [String, Fixnum, Bignum, NilClass] val
    # @return [Boolean] +true+ if the function succeeds.
    def write(path, val)
      path, name = format_path(path)
      hKeyPtr = 0.chr*4
      r = wins::RegCreateKey.call(HKEY_CURRENT_USER, path, hKeyPtr)
      return unless r.zero?
      hKey = hKeyPtr.unpack('L')[0]
      case val
      when Fixnum
        type = 4
        data = [val.abs].pack('L')
        size = 4
      when Bignum
        type = 11
        data = [val.abs].pack('Q')
        size = 8
      else
        type = 1
        data = (val.to_s+0.chr).unpack("U*").pack("C*") # Encode to ASCII-8BIT
        size = data.size
      end
      r = wins::RegSetValueEx.call(hKey, name, nil, type, data, size)
      wins::RegCloseKey.call(hKey)
      r.zero?
    end

    alias :set :write


    # Deletes the specified registry key or value.
    # @param [String] path
    # @return [Boolean] +true+ if removed, or +false+ if path is invalid.
    def delete(path)
      path, name = format_path(path)
      hKeyPtr = 0.chr*4
      r = wins::RegOpenKey.call(HKEY_CURRENT_USER, path, hKeyPtr)
      return false if r != 0
      hKey = hKeyPtr.unpack('L')[0]
      return true if wins::RegDeleteValue.call(hKey, name).zero?
      wins::RegDeleteKey.call(hKey, name).zero?
      wins::RegCloseKey.call(hKey)
    end

    alias :remove :delete


    # Get all 'folders' found in the specified registry path.
    # @param [String] path
    # @return [Array<String>] An array of folder names.
    def get_keys(path)
      path, name = format_path(path+'/x')
      hKeyPtr = 0.chr*4
      r = wins::RegOpenKey.call(HKEY_CURRENT_USER, path, hKeyPtr)
      return [] if r != 0
      hKey = hKeyPtr.unpack('L')[0]
      count = 0
      data = []
      while(true)
        buffer = 0.chr*256
        e = wins::RegEnumKey.call(hKey, count, buffer, buffer.size)
        break if e == 0x103 # ERROR_NO_MORE_ITEMS
        if e != 0xEA # ERROR_MORE_DATA
          name = buffer.unpack("C*").pack("U*").rstrip
          data.push(name)
        end
        count += 1
      end
      wins::RegCloseKey.call(hKey)
      data
    end


    # Get all 'files' and their values found in the specified registry path.
    # @param [String] path
    # @return [Hash] A hash of value names and their values.
    def get_values(path)
      path, name = format_path(path+'/x')
      hKeyPtr = 0.chr*4
      r = wins::RegOpenKey.call(HKEY_CURRENT_USER, path, hKeyPtr)
      return {} if r != 0
      hKey = hKeyPtr.unpack('L')[0]
      count = 0
      data = {}
      while(true)
        lpName = 0.chr*256
        lpNameSize = [lpName.size].pack('L')
        lpType = 0.chr*4
        lpData = 0.chr*2048
        lpDataSize = [lpData.size].pack('L')
        e = wins::RegEnumValue.call(hKey, count, lpName, lpNameSize, nil, lpType, lpData, lpDataSize)
        break if e == 0x103 # ERROR_NO_MORE_ITEMS
        if e != 0xEA # ERROR_MORE_DATA
          nameSize = lpNameSize.unpack('L')[0]
          name = lpName[0, nameSize].unpack("C*").pack("U*")
          type = lpType.unpack('L')[0]
          val = case type
          when 0 # REG_NONE
            nil
          when 1, 7 # REG_SZ, REG_MULTI_SZ
            dataSize = lpDataSize.unpack('L')[0]
            lpData[0, dataSize-1].unpack("C*").pack("U*")
          when 2 # REG_EXPAND_SZ
            buffer = 0.chr*2048
            s = wins::ExpandEnvironmentStrings.call(lpData, buffer, buffer.size)
            buffer[0, s-1].unpack("C*").pack("U*")
          when 3 # REG_BINARY
            lpData.unpack('A')[0]
          when 4 # REG_DWORD, REG_DWORD_LITTLE_ENDIAN
            lpData.unpack('L')[0]
          when 5 # REG_DWORD_BIG_ENDIAN
            lpData.unpack('q>')[0]
          when 6 # REG_LINK
            lpData.unpack('P')[0]
          when 11 # REG_QWORD, REG_QWORD_LITTLE_ENDIAN
            lpData.unpack('q')[0]
          end
          data[name] = val
        end
        count += 1
      end
      wins::RegCloseKey.call(hKey)
      data
    end

  end

  HKEY_CURRENT_USER = 0x80000001
  PATH = self.get_registry_path

end # module AMS::Sketchup::Registry

AMS::Lib.file_loaded(__FILE__)
