# ------------------------------------------------------------------------------
#
#   Anton Synytsia
#   anton.synytsia@gmail.com
#
# ------------------------------------------------------------------------------

Sketchup.require 'ams_Lib/sketchup_window.rb'

# Toolbar containers are the place-holders for the floating toolbars. SketchUp
# has four toolbar containers surrounding the rectangular viewport. They are
# also called control bars.
module AMS::Sketchup::ToolbarContainers

  class << self

    private

    def api
      AMS::Win32::API
    end

    def wins
      AMS::Windows
    end

  end # proxy class

  @rect = []
  @count = 0
  @filled = nil

  @enum_windows = api::Callback.new('LI', 'I'){ |hwnd, param|
    parent = AMS::Window.get_parent(hwnd)
    next true unless hwnds.include?(parent)
    style = wins::GetWindowLong.call(hwnd, -16)
    ws_visible = 0x10000000
    visible = ( style & ws_visible == ws_visible )
    case param
    when 1 # Computes container rect.
      if visible
        wr = AMS::Window.rect(hwnd, 1)
        @rect[0] = wr[0] if wr[0] < @rect[0]
        @rect[1] = wr[1] if wr[1] < @rect[1]
        @rect[2] = wr[2] if wr[2] > @rect[2]
        @rect[3] = wr[3] if wr[3] > @rect[3]
      end
      true
    when 2 # Hides docked toolbars.
      if visible
        new_style = style & ~ws_visible
        wins::SetWindowLong.call(hwnd, -16, new_style)
        @count += 1
      end
      true
    when 3 # Determines whether the container is filled.
      if visible
        @filled = true
        next false
      end
      true
    else
      false
    end
  }


  class << self

    # Retrieves an array of windows handles to all four toolbar containers of
    # the main window.
    # @return [Array] +[top, bottom, left, right]+
    def hwnds
      AMS::Sketchup.toolbar_containers
    end

    alias handles hwnds


    # Computes the new bounding rectangle of the specified toolbar container.
    # The new rect is based on the overall rect of its visbile child windows.
    # The coordinates are given relative to client origin of the main window.
    # @note This function is designed for SU 13+ only! Toolbar containers will
    # adjust automatically on SU 8.
    # @param [Integer] bar top (1), bottom (2), left (3), right (4), all (5).
    # @param [Boolean] state +true+ to adjust, +false+ to collapse.
    # @param [Boolean] refresh Whether to refresh after the operation.
    # @return [Boolean] +true+ if successful.
    def adjust_rect(bar = 5, state = true, refresh = true)
      if Sketchup.version.to_i < 13
        AMS::Sketchup.refresh if refresh == true
        return false
      end
      bar = bar.to_i
      bar = 5 unless bar.between?(1,5)
      state = (state == true)
      windows = ( bar == 5 ? hwnds : [hwnds[bar-1]] )
      o = AMS::Window.client_origin(AMS::Sketchup.main_window, 1)
      windows.each{ |handle|
        r = AMS::Window.rect(handle, 1)
        if state
          @rect = [r[2],r[3], r[0],r[1]]
          wins::EnumChildWindows.call(handle, @enum_windows, 1)
          case hwnds.index(handle)
          when 0; r[3] = @rect[3]
          when 1; r[1] = r[3] - (@rect[3] - @rect[1])
          when 2; r[2] = @rect[2]
          when 3; r[0] = r[2] - (@rect[2] - @rect[0])
          end
        else
          case hwnds.index(handle)
          when 0; r[3] = r[1]
          when 1; r[1] = r[3]
          when 2; r[2] = r[0]
          when 3; r[0] = r[2]
          end
        end
        r[0] -= o[0]
        r[1] -= o[1]
        r[2] -= o[0]
        r[3] -= o[1]
        wins::MoveWindow.call(handle, r[0], r[1], r[2]-r[0], r[3]-r[1], true)
      }
      AMS::Sketchup.refresh if refresh == true
      true
    end


    # Modifies the visibility state of the specified toolbar container(s).
    # @param [Boolean] state Whether to show or hide the window.
    # @param [Integer] bar top (1), bottom (2), left (3), right (4),
    #   all filled (5), all (6).
    # @param [Boolean] refresh Whether to refresh after the operation.
    # @return [Boolean] Whether visibility state of the +bar+ was changed.
    def show(bar = 6, state = true, refresh = true)
      bar = bar.to_i
      bar = 6 unless bar.between?(1,6)
      state = (state == true)
      cmd = ( state ? 4 : 0 )
      if bar == 5
        windows = []
        for n in 1..4
          windows.push(hwnds[n-1]) if filled?(n)
        end
      else
        windows = ( bar == 6 ? hwnds : [hwnds[bar-1]] )
      end
      origin = AMS::Window
      count = 0
      windows.each{ |handle|
        next if state == (wins::IsWindowVisible.call(handle) == 1)
        wins::ShowWindow.call(handle, cmd)
        adjust_rect(hwnds.index(handle)+1, state, false)
        count += 1
      }
      # Return true if at least ones visibility state changed; otherwise, false.
      return false if (count == 0)
      AMS::Sketchup.refresh if refresh == true
      true
    end


    # Removes docked toolbars from the specified toolbar container(s).
    # @param [Integer] bar top (1), bottom (2), left (3), right (4), all (5).
    # @param [Boolean] refresh Whether to refresh after the operation.
    # @return [Integer] The total number of docked toolbars removed.
    def clear(bar = 5, refresh = true)
      bar = bar.to_i
      bar = 5 unless bar.between?(1,5)
      windows = ( bar == 5 ? hwnds : [hwnds[bar-1]] )
      @count = 0
      windows.each { |handle|
        wins::EnumChildWindows.call(handle, @enum_windows, 2)
      }
      adjust_rect(5, false, refresh == true) if @count > 0
      @count
    end


    # Determines whether the specified toolbar container(s) is/are visible.
    # @param [Integer] bar top (1), bottom (2), left (3), right (4),
    #   all filled (5), all (6).
    # @return [Boolean] +true+ if visible, +false+ if not.
    def visible?(bar = 6)
      bar = bar.to_i
      bar = 6 unless bar.between?(1,6)
      if bar == 5
        windows = []
        for n in 1..4
          windows.push(hwnds[n-1]) if filled?(n)
        end
      else
        windows = ( bar == 6 ? hwnds : [hwnds[bar-1]] )
      end
      count = 0
      return false if windows.empty?
      windows.each{ |handle|
        count += 1 if wins::IsWindowVisible.call(handle) == 1
      }
      count == windows.size
    end


    # Determines whether the specified toolbar container(s) is/are populated.
    # @param [Integer] bar top (1), bottom (2), left (3), right (4),
    #   at least one (5), all (6).
    # @return [Boolean] +true+ if filled, +false+ if not.
    def filled?(bar = 6)
      bar = bar.to_i
      bar = 6 unless bar.between?(1,6)
      windows = ( bar > 4 ? hwnds : [hwnds[bar-1]] )
      windows.each { |handle|
        @filled = false
        wins::EnumChildWindows.call(handle, @enum_windows, 3)
        state = (bar == 6) ? !@filled : @filled
        break if state
      }
      @filled
    end


    # Determines whether the specified toolbar container(s) is/are empty.
    # @param [Integer] bar top (1), bottom (2), left (3), right (4),
    #   at least one (5), all (6).
    # @return [Boolean] +true+ if empty, +false+ if not.
    def empty?(bar = 6)
      bar = bar.to_i
      bar = 6 unless bar.between?(1,6)
      windows = ( bar > 4 ? hwnds : [hwnds[bar-1]] )
      windows.each { |handle|
        @filled = false
        wins::EnumChildWindows.call(handle, @enum_windows, 3)
        state = (bar == 5) ? !@filled : @filled
        break if state
      }
      !@filled
    end

  end # proxy class

end # module AMS::Sketchup::ToolbarContainers

AMS::Lib.file_loaded(__FILE__)
