# ------------------------------------------------------------------------------
#
#   Anton Synytsia
#   anton.synytsia@gmail.com
#
# ------------------------------------------------------------------------------

Sketchup.require 'ams_Lib/sketchup_window.rb'

# Viewport is the view window where all drawings take place.
module AMS::Sketchup::Viewport

  class << self

    private

    def wins
      AMS::Windows
    end

    public

    # Retrieves a handle to the main window's viewport.
    # @return [Long] hWnd
    def hwnd
      AMS::Sketchup.viewport
    end

    alias handle hwnd


    # Retrieves dimensions (upper-left and lower-right corners) of the window's
    # client area. The dimensions are given in screen coordinates that are
    # relative to the upper-left corner of the screen.
    # @return [Array] +[x1,y1, x2,y2]+
    def rect
      AMS::Window.client_rect(hwnd)
    end


    # Retrieves screen coordinates of the upper-left corner of the window.
    # @return [Array] +[x,y]+
    def origin
      rect[0..1]
    end


    # Retrieves window's size in pixels.
    # @return [Array] +[width,height]+
    def size
      r = rect
      w = r[2] - r[0]
      h = r[3] - r[1]
      [w, h]
    end


    # Retrieves window's width in pixels.
    # @return [Integer] width
    def width; size[0] end


    # Retrieves window's height in pixels.
    # @return [Integer] height
    def height; size[1] end


    # Retrieves window's center coordinates.
    # @return [Array] +[cx,cy]+
    def center
      p = rect
      cx = (p[0] + p[2]) / 2
      cy = (p[1] + p[3]) / 2
      [cx, cy]
    end


    # Retrieves an array of screen coordinates to all four corners (upper-left,
    # upper-right, lower-left, lower-right) of the view window.
    # @return [Array] +[[x1,y1], [x2,y2], [x3,y3], [x4,y4]]+
    def corners
      p = rect
      [[p[0],p[1]], [p[2],p[1]], [p[0],p[3]], [p[2],p[3]]]
    end


    # Sets or removes view client edge, a thin line border from the viewport
    # window.
    # @param [Boolean] state +true+ : set, +false+ : remove.
    # @return [Boolean] Whether window's state was changed.
    def set_client_edge(state = true)
      state = (state == true)
      exstyle = state ? 0x200 : 0
      return false if (state == client_edge_set?)
      wins::LockWindowUpdate.call hwnd
      wins::SetWindowLong.call(hwnd, -20, exstyle)
      wins::LockWindowUpdate.call nil
      wins::SetWindowPos.call(hwnd, 0, 0, 0, 0, 0, 0x0277)
      true
    end


    # Determines whether the viewport window has a thin line border.
    # @return [Boolean] +true+ if there, +false+ if not.
    def client_edge_set?
      exstyle = wins::GetWindowLong.call(hwnd, -20)
      ((exstyle | 0x00000200) == exstyle)
    end


    # Repaints the window.
    # @return [nil]
    def repaint
      wins::PostMessage.call(hwnd, 0x000F, 0, nil)
      hdc = wins::GetDC.call(hwnd)
      wins::DefWindowProc.call(hwnd, 0x0014, hdc, nil)
      wins::ReleaseDC.call(hwnd, hdc)
      nil
    end

  end # proxy class

end # module AMS::Sketchup::Viewport

AMS::Lib.file_loaded(__FILE__)
