# ------------------------------------------------------------------------------
#
#   Anton Synytsia
#   anton.synytsia@gmail.com
#
# ------------------------------------------------------------------------------

dir = File.dirname(__FILE__)
require File.join(dir, 'core.rb')

AMS::Lib.require 'window.rb'
AMS::Lib.require 'cursor.rb'

# This namespace is the base where all SketchUp handles and IDs are searched and
# stored. As well, it provides functions that rely to the current SketchUp
# application.
module AMS::Sketchup

  class << self

    private

    def api
      AMS::Win32::API
    end

    def wins
      AMS::Windows
    end

    def win
      AMS::Window
    end

  end # proxy class


  @pid = nil
  @tid = nil
  @hmod = wins::GetModuleHandle.call('SketchUp.exe')
  @path = nil
  @main_window = nil
  @menubar = nil
  @viewport = nil
  @containers = []
  @statusbar = nil
  @scenesbar = nil
  @toolbars = []
  @dialogs = []
  @dialogs_to_ignore = []
  @temp = { :hwnd => nil, :caption => nil }
  @hhook = []
  @active = false
  @observers = []
  @observer_methods = [
    :swo_mw_onCaptionChanged, :swo_mw_onSwitchFullscreen, :swo_mw_onMaximize,
    :swo_mw_onMinimize, :swo_mw_onRestore, :swo_mw_onActivate,
    :swo_mw_onDeactivate, :swo_mw_onEnterSizeMove, :swo_mw_onExitSizeMove,
    :swo_mw_onSizeMove, :swo_mw_onKeyDown, :swo_mw_onKeyExtended, :swo_mw_onKeyUp,
    :swo_mb_onSetRemove, :swo_mb_onVisibilityChanged, :swo_vp_onSetRemoveClientEdge,
    :swo_vp_onSize, :swo_vp_onMouseEnter, :swo_vp_onMouseLeave, :swo_vp_onRepaint,
    :swo_tbc_onVisibilityChanged, :swo_scb_onVisibilityChanged,
    :swo_stb_onVisibilityChanged, :swo_mw_onEnterMenu, :swo_mw_onExitMenu,
    :swo_mw_onCommand, :swo_tbc_onFilled, :swo_tbc_onEmptied
  ]
  @su_controlled_dialogs = [
    'Model Info', 'Entity Info', 'Materials', 'Components', 'Styles', 'Layers',
    'Outliner', 'Scenes', 'Shadow Settings', 'Fog', 'Match Photo',
    'Soften Edges', 'Instructor'
  ]
  @mb_set = true
  @mb_active = false
  @mb_state = false
  @vp_mouse = false
  @mw_placement = nil
  @call_delay = 0.02
  @tbc_status = []

  ver = ::Sketchup.version.to_i
  if ver < 13
    @cnames = {
      :container => (ver > 6 ? /AfxControlBar80u/i : /AfxControlBar70u/i),
      :scenesbar => /msctls_statusbar32/i,
      :statusbar => /msctls_statusbar32/i,
      :viewport => (ver > 6 ? 'AfxFrameOrView80u' : 'AfxFrameOrView70u'),
      :scene_tab => 'SysTabControl32',
      :docked_toolbar => /ToolbarWindow32/i,
      :floating_toolbar => %w(Afx 8 00000000),
      :dialog => '#32770'
    }
  else
    @cnames = {
      :container => /Afx:DockPane/i,
      :scenesbar => /msctls_statusbar32/i,
      :statusbar => /Afx:RibbonStatusBar/i,
      :viewport => 'AfxFrameOrView100u',
      :scene_tab => 'SysTabControl32',
      :docked_toolbar => /Afx:ToolBar/i,
      :floating_toolbar => /Afx:MiniFrame/i,
      :dialog => '#32770'
    }
  end

  # Use to find handle to the the main window.
  @find_main_window = api::Callback.new('LP', 'I'){ |hwnd, param|
    # Get thread and process IDs of current enumerated hWnd.
    struct = 0.chr*4
    cur_tid = wins::GetWindowThreadProcessId.call(hwnd, struct)
    cur_pid = struct.unpack('l')[0]
    # Skip the rest operation if pid of current enumerated window does not match
    # current pid.
    next true unless cur_pid == @pid
    # Check for the match in class name.
    cname = 0.chr*50
    wins::GetClassName.call(hwnd, cname, cname.size)
    cname.strip!
    # SU class name sample:
    #   Afx:010A0000:b:00010003:00000006:27E00BDD
    # SketchUp class name is not constant, only odd parts, separated by ":"
    # remain constant.
    if [cname[0,3], cname[13,1], cname[24,8]] == %w(Afx b 00000006)
      @tid = cur_tid
      @main_window = hwnd
      next false
    end
    true
  }

  # Use to gathers all child windows.
  @find_child_windows = api::Callback.new('LP', 'I'){ |hwnd, param|
    cname = 0.chr*50
    wins::GetClassName.call(hwnd, cname, cname.size)
    cname.strip!
    if cname =~ @cnames[:container]
      @containers.push(hwnd)
    elsif cname =~ @cnames[:scenesbar]
      if Object::Sketchup.version.to_i < 13
        chwnd = wins::FindWindowEx.call(hwnd, nil, @cnames[:scene_tab], nil)
        (chwnd == 0) ? @statusbar = hwnd : @scenesbar = hwnd
      else
        @scenesbar = hwnd
      end
    elsif cname =~ @cnames[:statusbar]
      @statusbar = hwnd
    elsif cname == @cnames[:viewport]
      @viewport = hwnd
    end
    true
  }

  # Use to gather floating toolbars/dialogs, or find window by caption.
  @enum_windows = api::Callback.new('LI', 'I'){ |hwnd, param|
    if param == 0 # find floating toolbars
      cname = 0.chr*50
      wins::GetClassName.call(hwnd, cname, cname.size)
      cname.strip!
      if (Object::Sketchup.version.to_i < 13)
        match = ([cname[0,3], cname[13,1], cname[24,8]] == @cnames[:floating_toolbar])
      else
        match = (cname =~ @cnames[:floating_toolbar])
      end
      @toolbars.push(hwnd) if ( match and (wins::IsWindowVisible.call(hwnd) == 1) and !@toolbars.include?(hwnd) )
      true
    elsif param == 1 # find dialogs
      cname = 0.chr*50
      wins::GetClassName.call(hwnd, cname, cname.size)
      cname.strip!
      match = ( (cname == @cnames[:dialog]) and win.visible?(hwnd) and !@dialogs.include?(hwnd) )
      @dialogs.push(hwnd) if match
      true
    elsif param == 2 # find window by caption
      if win.get_caption(hwnd) == @temp[:caption]
        @temp[:hwnd] = hwnd
        false
      else
        true
      end
    else
      false
    end
  }

  # GetMsgProc (3)
  @msg_proc = api::Callback.new('ILI', 'I'){ |nCode, param1, param2|
    if nCode != 0 or param1 == 0
      next wins::CallNextHookEx.call(@hhook[0], nCode, param1, param2)
    end
    buffer = 0.chr*28
    wins::Memcpy2.call(buffer, param2, buffer.size)
    hwnd, msg, wParam, lParam, time, x, y = buffer.unpack("LLllLll")
    method = nil
    params = []
    case hwnd
    when @main_window
      if msg == 0x0111 # WM_COMMAND
        method = :swo_mw_onCommand
        params = [wParam]
      end
    when viewport
      case msg
      when 0x0100, 0x0101, 0x0104, 0x0105 # WM_KEYDOWN, WM_KEYUP, WM_SYSKEYDOWN, WM_SYSKEYUP
        unless @mb_active
          if [0x0100, 0x0104].include?(msg)
            method = (lParam & (1 << 30)) == 0 ? :swo_mw_onKeyDown : :swo_mw_onKeyExtended
          else
            method = :swo_mw_onKeyUp
          end
          # Associate the message value with its name.
          func = RUBY_VERSION.to_f < 1.9 ? :index : :key
          vk = wins::VirtualKeyCodes.method(func).call(wParam).to_s.downcase
          params = [vk, wParam]
        end
      when 0x0200 # WM_MOUSEMOVE
        unless @vp_mouse
          method = :swo_vp_onMouseEnter
          @vp_mouse = true
        end
      when 0x02A3 # WM_MOUSELEAVE
        method = :swo_vp_onMouseLeave
        @vp_mouse = false
      when 0x000f # WM_PAINT
        method = :swo_vp_onRepaint
      end
    end
    if method
      # Call the method associated with the message, after the window procedure
      # finishes processing the message. This way we will avoid calling
      # "on before" events.
      UI.start_timer(@call_delay, false){
        method = method.to_s if RUBY_VERSION.to_f < 1.9
        @observers.each { |obs|
          next unless obs.methods.include?(method)
          begin
            obs.method(method).call *params
          rescue Exception => e
            puts "There was an error calling #{obs}.#{method} method:\n#{e}\n#{$@[0..2].join("\n")}"
            remove_observer(obs)
          end
        }
      }
    end
    wins::CallNextHookEx.call(@hhook[0], nCode, param1, param2)
  }

  # CallWndProc (4) / CallWndRetProc (12)
  @wnd_proc = api::Callback.new('ILI', 'I'){ |nCode, param1, param2|
    if nCode != 0
      next wins::CallNextHookEx.call(@hhook[1], nCode, param1, param2)
    end
    buffer = 0.chr*20
    wins::Memcpy2.call(buffer, param2, buffer.size)
    #lResult, lParam, wParam, msg, hwnd = buffer.unpack("lllLL")
    lParam, wParam, msg, hwnd = buffer.unpack("llLL")
    method = nil
    params = []
    case hwnd
    when @main_window
      case msg
      when 0x0006 # WM_ACTIVATE
        method = ( wParam == 0 ? :swo_mw_onDeactivate : :swo_mw_onActivate )
      when 0x000C # WM_SETTEXT
        method = :swo_mw_onCaptionChanged
        struct = 0.chr*256
        wins::Memcpy2.call(struct, lParam, struct.size)
        text = (Sketchup.version.to_i < 14 ? struct.split(0.chr) : struct.unpack("Z*"))[0]
        params = [text]
      when 0x0211 # WM_ENTERMENULOOP
        @mb_active = true
        method = :swo_mw_onEnterMenu
      when 0x0212 # WM_EXITMENULOOP
        @mb_active = false
        method = :swo_mw_onExitMenu
      when 0x0005 # WM_SIZE
        case wParam
        when 0
          if @mw_placement != 0
            if @mw_placement == 3
              method = :swo_mw_onSwitchFullscreen
              params = [false]
            else
              method = :swo_mw_onRestore
            end
            @mw_placement = 0
          end
        when 1
          if @mw_placement != 1
            method = :swo_mw_onMinimize
            @mw_placement = 1
          end
        when 2
          if MainWindow.fullscreen?
            method = :swo_mw_onSwitchFullscreen
            params = [true]
            @mw_placement = 3
          else
            if @mw_placement == 3
              method = :swo_mw_onSwitchFullscreen
              params = [false]
            elsif @mw_placement != 2
              method = :swo_mw_onMaximize
            end
            @mw_placement = 2
          end
        end
      when 0x0231 # WM_ENTERSIZEMOVE
        method = :swo_mw_onEnterSizeMove
      when 0x0232 # WM_EXITSIZEMOVE
        method = :swo_mw_onExitSizeMove
      when 0x0214, 0x0216 # WM_SIZING, WM_MOVING
        method = :swo_mw_onSizeMove
        struct = 0.chr*16
        wins::Memcpy2.call(struct, lParam, struct.size)
        params = struct.unpack('llll')
      when 0x0047 # WM_WINDOWPOSCHANGED
        # Use this message to determine whether menubar has been set or removed.
        # I haven't found the exact notification that would react to menubar
        # being set or removed.
        state = nil
        if wins::GetMenu.call(hwnd) == @menubar and !@mb_set
          state = true
        elsif wins::GetMenu.call(hwnd) != @menubar and @mb_set
          state = false
        end
        if state == nil
          # Determine whether the menubar visibility state was changed.
          state = AMS::Sketchup::Menubar.visible?
          if state != @mb_state
            @mb_state = state
            method = :swo_mb_onVisibilityChanged
            params = [state]
          end
        else
          @mb_set = state
          method = :swo_mb_onSetRemove
          params = [state]
        end
      end
    when viewport
      case msg
      when 0x007D # WM_STYLECHANGED
        if wParam == -20
          struct = 0.chr*8
          wins::Memcpy2.call(struct, lParam, struct.size)
          style1, style2 = struct.unpack('ll')
          state = nil
          v = 0x00000200
          if (style1 & v) == 0 and (style2 & v) == v
            state = true
          elsif (style1 & v) == v and (style2 & v) == 0
            state = false
          end
          if state != nil
            method = :swo_vp_onSetRemoveClientEdge
            params = [state]
          end
        end
      when 0x0005 # WM_SIZE
        method = :swo_vp_onSize
        w = lParam & 0xffff
        h = lParam >> 16
        params = [w,h]
      when 0x000f # WM_PAINT
        method = :swo_vp_onRepaint
      end
    when @containers[0], @containers[1], @containers[2], @containers[3], @scenesbar, @statusbar
      case msg
      when 0x0005 # WM_SIZE
        if @containers.include?(hwnd)
          w = lParam & 0xFFFF
          h = lParam >> 16
          s = (w < 1 or h < 1)
          i = @containers.index(hwnd)
          if s != @tbc_status[i]
            method = s ? :swo_tbc_onEmptied : :swo_tbc_onFilled
            params = [i+1]
            @tbc_status[i] = s
          end
        end
      when 0x0047 # WM_WINDOWPOSCHANGED
        struct = 0.chr*28
        wins::Memcpy2.call(struct, lParam, struct.size)
        h, ha, x, y, cx, cy, flags = struct.unpack('LLllllL')
        if flags & 0x0040 == 0x0040 or flags & 0x0080 == 0x0080
          params = [flags & 0x0040 == 0x0040]
          case hwnd
          when @scenesbar
            method = :swo_scb_onVisibilityChanged
          when @statusbar
            method = :swo_stb_onVisibilityChanged
          else
            method = :swo_tbc_onVisibilityChanged
            bar = @containers.index(hwnd)+1
            params.insert(0, bar)
          end
        end
      when 0x007D # WM_STYLECHANGED
        if wParam == -16 and hwnd == @statusbar
          struct = 0.chr*8
          wins::Memcpy2.call(struct, lParam, struct.size)
          style1, style2 = struct.unpack('ll')
          state = nil
          v = 0x10000000
          if (style1 & v) == 0 and (style2 & v) == v
            state = true
          elsif (style1 & v) == v and (style2 & v) == 0
            state = false
          end
          if state != nil
            method = :swo_stb_onVisibilityChanged
            params = [state]
          end
        end
      end
    end
    if method
      # Call the method associated with the message, after the window procedure
      # finishes processing the message. This way we will avoid calling
      # "on before" events.
      UI.start_timer(@call_delay, false){
        method = method.to_s if RUBY_VERSION.to_f < 1.9
        @observers.each { |obs|
          next unless obs.methods.include?(method)
          begin
            obs.method(method).call *params
          rescue Exception => e
            puts "There was an error calling #{obs}.#{method} method:\n#{e}\n#{$@[0..2].join("\n")}"
            remove_observer(obs)
          end
        }
      }
    end
    wins::CallNextHookEx.call(@hhook[1], nCode, param1, param2)
  }


  class << self

    # Retrieves a hash of class names that are part of SketchUp window.
    # @return [Hash]
    def class_names; @cnames.dup end


    # Retrieves an identifier to the process that created SketchUp window.
    # @return [Long] process id
    def process_id; @pid end

    alias pid process_id


    # Retrieves an identifier to the thread that created SketchUp window.
    # @return [Long] thread id
    def thread_id; @tid end

    alias tid thread_id


    # Retrieves module handle of SketchUp window module.
    # @return [Long] hMod
    def module_handle; @hmod end

    alias hmod module_handle


    # Retrieves a handle to the top, parent window of SketchUp application.
    # @return [Long] hWnd
    def main_window; @main_window end


    # Retrieves a handle to the main window's menubar.
    # @return [Long] hMenu
    def menubar; @menubar end


    # Retrieves an array of window handles to all four toolbar containers of the
    # main window.
    # @return [Array<Long>] +[top, bottom, left, right]+
    def toolbar_containers; @containers.dup end


    # Retrieves a window handle to the main window's statusbar.
    # @return [Long] hWnd
    def statusbar; @statusbar end


    # Retrieves a window handle to the main window's scenesbar.
    # @return [Long] hWnd
    def scenesbar; @scenesbar end


    # Retrieves a window handle to the main window's viewport.
    # @return [Long] hWnd
    def viewport
      if wins::IsWindow.call(@viewport) == 0
        @viewport = wins::FindWindowEx.call(@main_window, nil, @cnames[:viewport], nil)
      end
      @viewport
    end


    attr_reader :dialogs
    attr_reader :toolbars

    # Determines whether a foreground window is part of the current SketchUp
    # application.
    # @return [Boolean]
    def active?
      hwnd = wins::GetForegroundWindow.call()
      # Get thread and process id of current enumerated window hWnd.
      struct = 0.chr*4
      cur_tid = wins::GetWindowThreadProcessId.call(hwnd, struct)
      cur_pid = struct.unpack('l')[0]
      cur_tid == @tid
    end


    # Attempts to find a handle to the main window.
    # @param [Long] cur_pid An identifier to the process.
    # @return [Boolean] +true+ if found, +false+ if not.
    def find_main_window(cur_pid = Object::Process.pid)
      @pid = cur_pid
      @main_window = nil
      wins::EnumWindows.call(@find_main_window, nil)
      (@main_window != nil)
    end


    # Searches for window handles to all four toolbar containers, statusbar,
    # scenesbar, and viewport.
    # @return [Boolean] +true+ if gathered, +false+ if a handle to the main
    #   window is undefined.
    def find_child_windows
      return false unless @main_window
      @menubar = wins::GetMenu.call(@main_window)
      wins::EnumChildWindows.call(@main_window, @find_child_windows, nil)
      true
    end


    # Retrieves a handle to the window with the specified title text. This method
    # enumerates all windows part of SketchUp thread, and returns the first window
    # with the specified caption.
    # @param [String] text A title text to the window.
    # @return [Long, nil] hWnd, or +nil+ if not found.
    def window_by_caption(text)
      @temp = { :hwnd => nil, :caption => text }
      wins::EnumThreadWindows.call(@tid, @enum_windows, 2)
      @temp[:hwnd]
    end


    # Retrieves the number of active floating toolbars. These include all
    # visible floating toolbars and those hidden by the {show_toolbars} method.
    # @param [Boolean] find Whether to regather floating toolbars.
    # @return [Integer] count
    def get_toolbars(find = true)
      if find == true
        wins::EnumThreadWindows.call(@tid, @enum_windows, 0)
        # Remove invalid toolbars
        @toolbars.dup.each{ |hwnd|
          @toolbars.delete(hwnd) if wins::IsWindow.call(hwnd) == 0
        }
      end
      @toolbars.size
    end


    # Retrieves the number of all visible flaoting toolbars.
    # @param [Boolean] find Whether to regather floating toolbars.
    # @return [Integer] count
    def get_visible_toolbars(find = true)
      get_toolbars if find == true
      count = 0
      @toolbars.each{ |hwnd|
        count += 1 if win.visible?(hwnd)
      }
      count
    end


    # Shows or hides floating toolbars, the toolbars that are not embedded into
    # the toolbar containers.
    # @param [Boolean] state +true+ to show, +false+ to hide.
    # @return [Boolean] +true+ if successful, +false+ if a handle to the main
    #   window is undefined, or there is no active floating toolbars.
    def show_toolbars(state = true)
      return false unless @main_window
      count = get_toolbars
      return false if count.zero?
      cmd = ( state == true ? 5 : 0 )
      @toolbars.each{ |hwnd| wins::ShowWindow.call(hwnd, cmd) if win.visible?(hwnd) != state }
      true
    end


    # Retrieves the number of active dialogs. These include all visible dialogs
    # and those hidden by the {show_dialogs} method, but not the ignored dialogs
    # that were ignored by the #{ignore_dialog} method.
    # @param [Boolean] find Whether to regather dialogs.
    # @return [Integer] count
    def get_dialogs(find = true)
      if find == true
        wins::EnumThreadWindows.call(@tid, @enum_windows, 1)
        # Remove invalid dialogs
        @dialogs_to_ignore.dup.each{ |hwnd|
          @dialogs_to_ignore.delete(hwnd) if wins::IsWindow.call(hwnd) == 0
        }
        @dialogs.dup.each{ |hwnd|
          @dialogs.delete(hwnd) if wins::IsWindow.call(hwnd) == 0 or @dialogs_to_ignore.include?(hwnd)
        }
      end
      @dialogs.size
    end


    # Retrieves the number of all visible dialogs, regarding the ignored dialgos.
    # @param [Boolean] find Whether to regather dialogs.
    # @return [Integer] count
    def get_visible_dialogs(find = true)
      get_dialogs if find == true
      count = 0
      @dialogs.each{ |hwnd|
        count += 1 if win.visible?(hwnd)
      }
      count
    end


    # Shows or hides all dialogs, including the Ruby Console.
    # @param [Boolean] state +true+ to show, +false+ to hide.
    # @return [Boolean] +true+ if successful, +false+ if a handle to the main
    #   window is undefined, or there is no active dialogs.
    def show_dialogs(state = true)
      return false unless @main_window
      state = ( state == true )
      count = get_dialogs
      return false if count.zero?
      cmd = state ? 4 : 0
      item = state ? 'Show Dialogs' : 'Hide Dialogs'
      path = "#{Menubar.locale2['Window']}/#{Menubar.locale.GetString(item)}"
      id = Menubar.get_menu_id_by_path(path)
      ::Sketchup.send_action(id) if id
      found = false
      UI.start_timer(0.1, false){
        @dialogs.each{ |hwnd|
          next if win.visible?(hwnd) == state
          if @su_controlled_dialogs.include?(win.get_caption(hwnd))
            found = true
            next
          end
          wins::ShowWindow.call(hwnd, cmd)
        }
        @dialogs.clear if state
        if found
          id = Menubar.get_menu_id_by_path("#{Menubar.locale2['Window']}/#{Menubar.locale.GetString('Show Dialogs')}")
          id = Menubar.get_menu_id_by_path("#{Menubar.locale2['Window']}/#{Menubar.locale.GetString('Hide Dialogs')}") unless id
          2.times { ::Sketchup.send_action(id) } if id
        end
      }
      true
    end


    # Eludes the specified dialog from the {show_dialogs} and {close_dialogs}
    # operations.
    # @param [Long] hwnd A handle to the window to ignore.
    # @return [Boolean] +true+ if successful, +false+ if a handle to the main
    #   window is undefined or the specified parameter does not meet demands of
    #   being a SketchUp dialog.
    def ignore_dialog(hwnd)
      return false unless @main_window
      hwnd = hwnd.to_i
      return false if @dialogs_to_ignore.include?(hwnd)
      # Verify given hwnd is a window
      return false unless wins::IsWindow.call(hwnd) == 1
      # Verify given hwnd is a dialog
      cname = 0.chr*50
      wins::GetClassName.call(hwnd, cname, cname.size)
      cname.strip!
      return false unless cname == '#32770'
      # Verify given hwnd belongs to current thread
      id = wins::GetWindowThreadProcessId.call(hwnd, nil)
      return false unless id == @tid
      # Push into queue once the hwnd met all the demands
      @dialogs_to_ignore.push(hwnd)
      true
    end


    # Includes the specified dialog in the {show_dialogs} operation. By default
    # all dialogs are included. This method is used to remove a dialog from the
    # to ignore list.
    # @param [Integer] hwnd A handle to the window to include.
    # @return [Boolean] +true+ if successful, +false+ if not.
    def include_dialog(hwnd)
      return false unless @main_window
      return false unless @dialogs_to_ignore.include?(hwnd)
      @dialogs_to_ignore.delete(hwnd)
      true
    end


    # Closes all visible floating toolbars, and floating toolbars that were
    # hidden by the {show_toolbars} method.
    # @return [Boolean] +true+ if successful, +false+ if a handle to the main
    #   window is undefined, or there is no active floating toolbars.
    def close_toolbars
      return false unless @main_window
      count = get_toolbars
      return false if count.zero?
      @toolbars.each{ |hwnd|
        #wins::SendMessage.call(hwnd, 0x0010, nil, nil) # WM_CLOSE
        wins::SendMessage.call(hwnd, 0x0112, 0xf060, 0) # WM_SYSCOMMAND
      }
      @toolbars.clear
      true
    end


    # Closes all visible dialogs, and dialogs that were hidden by the
    # {show_dialogs} method, but not the ignored dialogs that were ignored by
    # the #{ignore_dialog} method.
    # @return [Boolean] +true+ if successful, +false+ if a handle to the main
    #   window is undefined, or there is no active dialogs.
    def close_dialogs
      return false unless @main_window
      count = get_dialogs
      return false if count.zero?
      @dialogs.each{ |hwnd|
        #wins::SendMessage.call(hwnd, 0x0010, nil, nil) # WM_CLOSE
        wins::SendMessage.call(hwnd, 0x0112, 0xf060, 0) # WM_SYSCOMMAND
      }
      @dialogs.clear
      true
    end


    # Repositions and resizes all SketchUp windows of the current SketchUp
    # application. This includes web-dialogs, and child windows.
    # @note This method calls <i>WM_SIZE</i> message, which is not the most
    #   appropriate way to refresh the window. Improvements to this method are
    #   attempted!
    # @return [Boolean] +true+ if the function succeeds.
    def refresh
      return false unless @main_window
      wins::SendMessage.call(@main_window, 0x0005, 0, nil) # WM_SIZE
      #wins::PostMessage.call(@main_window, 0xc08d, 0, nil) # AFX_WM_CHANGEVISUALMANAGER
      true
    end


    # Sent as a signal that an application should terminate. This function does
    # the same operation as clicking the close button.
    # @return [Boolean] +true+ if the function succeeds.
    def close
      return false unless @main_window
      wins::SendMessage.call(@main_window, 0x0010, nil, nil) == 1
    end


    # Forcibly closes the current SketchUp application.
    # @return [Boolean] +true+ if the function succeeds.
    def end_task
      return false unless @main_window
      wins::EndTask.call(@main_window, false, true) == 1
    end


    # Activates observers.
    # @return [Boolean] Whether state changed.
    def activate_observers
      return false if @active or @observers.empty?
      @hhook[0] = wins::SetWindowsHookEx.call(3, @msg_proc, nil, @tid)
      @hhook[1] = wins::SetWindowsHookEx.call(4, @wnd_proc, nil, @tid)
      @active = true
      @mb_set = Menubar.set?
      @mb_state = Menubar.visible?
      @vp_mouse = AMS::Cursor.viewport_target? and MainWindow.active?
      @mw_placement = (MainWindow.fullscreen? ? 3 : (MainWindow.maximized? ? 2 : (MainWindow.minimized? ? 1 : 0)))
      @tbc_status.clear
      for i in 1..4; @tbc_status[i-1] = ToolbarContainers.filled?(i) end
      true
    end


    # Deactivates observers.
    # @return [Boolean] Whether state changed.
    def deactivate_observers
      return false unless @active
      @hhook.each{ |hk| wins::UnhookWindowsHookEx.call(hk) }
      @hhook.clear
      @active = false
      true
    end


    # Specifies whether the observers are active.
    # @return [Boolean]
    def observers_active?
      @active
    end


    # Attaches the specified module, class, or class instance to the observers
    # list. The specified object must contain one or more of the observer
    # callback methods, regarding the +swo_activate+ and +swo_deactivate+;
    # otherwise, the specified observer won't be added to the list.
    # @param [Object] obj A module, class, or class instance containing at least
    #   one of the SketchUp observer methods.
    # @return [Boolean] +true+ if successful, +false+ if not.
    # @see AMS::Sketchup::Observer
    def add_observer(obj)
      return false if @observers.include?(obj)
      methods = obj.methods
      # Convert strings to symbols
      for n in 0...methods.size
        methods[n] = methods[n].to_sym
      end if RUBY_VERSION.to_f < 1.9
      return false if (@observer_methods & methods).empty?
      method = :swo_activate
      method = method.to_s if RUBY_VERSION.to_f < 1.9
      begin
        obj.method(method).call
      rescue Exception => e
        puts "There was an error calling #{obj}.#{method} method:\n#{e}\n#{$@[0..2].join("\n")}"
        return false
      end if obj.methods.include?(method)
      @observers.push(obj)
      activate_observers
      true
    end


    # Removes the specified module, class, or class instance from the observers
    # list.
    # @param [Object] obj A module, class, or class instance that was once
    #   attached to the observers list.
    # @return [Boolean] +true+ if successful, +false+ if not.
    def remove_observer(obj)
      return false unless @observers.include?(obj)
      @observers.delete(obj)
      method = :swo_deactivate
      method = method.to_s if RUBY_VERSION.to_f < 1.9
      begin
        obj.method(method).call
      rescue Exception => e
        puts "There was an error calling #{obj}.#{method} method:\n#{e}\n#{$@[0..2].join("\n")}"
      end if obj.methods.include?(method)
      deactivate_observers if @observers.empty?
      true
    end


    # Get location of sketchup executable.
    # @return [String]
    def path
      return @path if @path
      buffer = 0.chr*1024
      wins::GetModuleFileName.call(@hmod, buffer, buffer.size)
      @path = File.dirname(buffer.strip.gsub('\\', '/'))
    end

  end # proxy class


  # @private
  class MyAppObserver < ::Sketchup::AppObserver

    def onQuit
      # Ensures safe unhook.
      ::AMS::Sketchup.deactivate_observers
      # Make sure status bar is visible on the next launch.
      if Sketchup::version.to_i > 8
        ::AMS::Sketchup::Registry.write("Workspace/BasePane-59393/IsVisible", 1)
      else
        count = 0
        while(true)
          id = ::AMS::Sketchup::Registry.read("ToolbarsUser-Bar#{count}/BarID")
          break if id.nil?
          if id == 59393
            ::AMS::Sketchup::Registry.write("ToolbarsUser-Bar#{count}/Visible", 1)
          end
          count += 1
        end
      end
      # Set menu bar to avoid getting post errors.
      ::AMS::Sketchup::Menubar.set
    end

  end # class MyAppObserver


  unless AMS::Lib.file_loaded?(__FILE__)
    # Require all sub-files.
    dir = File.dirname(__FILE__)
    files = Dir.glob(File.join(dir, 'SketchUp', '*.{rb, rbs}'))
    files.each{ |file| AMS::Lib.require file }
    # Find all windows and sub-windows.
    find_main_window
    find_child_windows
    # Add Application observer.
    ::Sketchup.add_observer(MyAppObserver.new)
    # Run custom menu observer.
    Menubar.activate_custom_menu
    ::AMS::Lib.file_loaded(__FILE__)
  end

end # module AMS::Sketchup
